/*
 *  MSP430F2122 Demo - ADC10, I2C Sample Code for bq76925 Analog Front End (AFE)
 *
 *  Description: This sample code demonstrates I2C communication and A/D
 *  conversion in a system comprising the bq76925 and an MSP430. The I2C 
 *  communication examples show how to read and write the bq76925 registers.
 *  The A/D conversion examples demonstrate the necessary procedures to obtain
 *  current, voltage and temperature readings from the VIOUT, VCOUT and VTB 
 *  pins of the bq76925.
 *
 *  The following connections between the bq76925 and MSP430 are assumed:
 *
 *  | bq76925 pin name | MSP430 pin name |
 *  | ---------------- | --------------- |
 *  |        SCL       |     UCB0SCL     |
 *  |        SDA       |     UCB0SDA     |
 *  |        VTB       |        A0*      |
 *  |       VCOUT      |        A1       |
 *  |       VIOUT      |        A2       |
 *
 *  * VTB is not connected directly to A0. It is connected through a series
 *  resistance that makes up part of a thermistor network as follows:
 *
 *  VTB -----[ 10K ]----- *----- A0
 *                       |
 *                       |
 *                     103AT
 *                       |
 *                       |
 *                      VSS
 *
 *  Michael Lu / J. Ross
 *  Texas Instruments Inc.
 *  Rev 1.1, 2 Sept 2011
 *  Built with CCS Version: 4.5.0
 *
 *  Revision history:
 *
 *      Date        Rev     Description
 *  ------------    ---     ---------------------------------------------------
 *   6 July 2011    1.0     Initial release
 *   2 Sept 2011    1.1     Changed thermistor to 103AT
 *
 * Copyright (C) 2011 Texas Instruments Incorporated - http://www.ti.com/ 
 * 
 * 
 *  Redistribution and use in source and binary forms, with or without 
 *  modification, are permitted provided that the following conditions 
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the   
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
*/

// =============================================================================
// Include header files
// =============================================================================
#include <intrinsics.h>
#include "msp430x21x2.h"

// =============================================================================
// Macro definitions
// =============================================================================
// MSP430 watchdog and delay macros
#define HALT_WATCHDOG   WDTCTL = WDTPW + WDTHOLD
#define RESET_WATCHDOG  WDTCTL = WDTPW + WDTCNTCL + WDTSSEL + WDTIS0
#define DELAY_32MS __delay_cycles(256000)

// bq76925 I2C register address macros
#define I2C_GROUP_ADDR  0x04
// Register addresses 0x00 - 0x05 are Read / Write.
// Register addresses 0x07 and 0x10 - 0x16 are read only. 
#define STATUS          0x00
#define CELL_CTL        0x01
#define BAL_CTL         0x02
#define CONFIG_1        0x03
#define CONFIG_2        0x04
#define POWER_CTL       0x05
#define CHIP_ID         0x07
#define VREF_CAL        0x10
#define VC1_CAL         0x11
#define VC2_CAL         0x12
#define VC3_CAL         0x13
#define VC4_CAL         0x14
#define VC5_CAL         0x15
#define VC6_CAL         0x16

// =============================================================================
// Type definitions
// =============================================================================
typedef struct{                     // Create a structure that will hold the 
        unsigned char data;         // most recent received or transmitted data
        unsigned char error;              
} i2c_trans;

// =============================================================================
// Function prototypes
// =============================================================================
void clock_init (void);

void i2c_init (void);

int i2c_read (unsigned char reg_addr, i2c_trans *i2c_read);

int i2c_write (unsigned char reg_addr, unsigned char data, i2c_trans *i2c_write);

void ADC_init (void);

unsigned int ADC_read (unsigned int ch_selected);

void error_trap (void);

// =============================================================================
// Functions
// =============================================================================
int main (void) {
    // -------------------------------------------------------------------------
    // Main loop
    // -------------------------------------------------------------------------
 
    // Initialize structures to hold latest read and write transaction
    i2c_trans i2c_read_struct = {0,0};
    i2c_trans i2c_write_struct = {0,0};
    
    // ADC results storage
    unsigned int adc_data = 0;      // ADC data from conversion
    unsigned int adc_sensen = 0;    // Store value of SENSEN current measurement

    // Loop index
    int index = 0;
    
    // MSP430 configuration setup
    clock_init();                   // Initialize clock system
    i2c_init();                     // Initialize I2C module
    ADC_init();                     // Initialize ADC10 module
    
    // bq76925 configuration setup.
    // Write CONFIG_1 register to select current measurement of SENSEN and
    // a gain of 4.
    if (i2c_write(CONFIG_1, 0x00, &i2c_write_struct)) {
        error_trap();
    }
    // Write CONFIG_2 register to select VREF = 3.0 V
    if (i2c_write(CONFIG_2, 0x01, &i2c_write_struct)) {
        error_trap();
    }
    // Write POWER_CTL register to enable VREF, VCOUT and VIOUT.
    if (i2c_write(POWER_CTL, 0x0D, &i2c_write_struct)) {
        error_trap();
    }

    // Verify bq76925 configuration
    // Verify CONFIG_1 register
    if (i2c_read(CONFIG_1, &i2c_read_struct) || i2c_read_struct.data != 0x00) {
        error_trap();
    };
    // Verify CONFIG_2 register
    if (i2c_read(CONFIG_2, &i2c_read_struct) || i2c_read_struct.data != 0x01) {
        error_trap();
    };
    // Verify POWER_CTL register
    if (i2c_read(POWER_CTL, &i2c_read_struct) || i2c_read_struct.data != 0x0D) {
        error_trap();
    }
    
    // Continuous measurement of VIOUT, VCOUT, and VTB.
    //
    // Eight AFE channels are defined as shown in the table below. The channels
    // are cycled through continuously to measure current via VIOUT, all cell
    // voltages (VC1 - VC6) via VCOUT, and temperature from a thermistor biased
    // from VTB.
    //
    // |             |  AFE Cell  |  Measure   | MSP430 Analog |
    // | AFE Channel | MUX Select | AFE Output | Input Select  |
    // | ----------- | ---------- | ---------- | ------------- |
    // |      0      |    VSS     |    VTB     |      A0       |
    // |      1      |    VC1     |   VCOUT    |      A1       |
    // |      2      |    VC2     |   VCOUT    |      A1       |
    // |      3      |    VC3     |   VCOUT    |      A1       |
    // |      4      |    VC4     |   VCOUT    |      A1       |
    // |      5      |    VC5     |   VCOUT    |      A1       |
    // |      6      |    VC6     |   VCOUT    |      A1       |
    // |      7      |    VSS     |   VIOUT    |      A2       |
    //
    // ------------------------------------------------------------------------
    // NOTE: This code DOES NOT apply the VREF and VCn correction factors to
    //       the ADC results. For best measurement accuracy, the correction
    //       factors must be read out and applied to the ADC results. See the
    //       bq76925 datasheet for details.
    // ------------------------------------------------------------------------

    // Before starting loop, the SENSEN current measurement must be
    // calibrated. Delay 32 ms before measurement.
    DELAY_32MS;
    adc_sensen = ADC_read(INCH_2);

    // Switch bq76925 to SENSEP for all other current measurements
    if (i2c_write(CONFIG_1, 0x04, &i2c_write_struct)) {
        error_trap();
    }

    // Debug code start
    // The P1.0 GPIO will be toggled each time through loop
    P1DIR = 0x01;
    P1OUT = 0x00;
    // Debug code end

    while (1) {
        // Cycle through all the channels as listed in the table
        for (index = 0; index < 8; index++) {
        
            if (index == 0) {
                // Measure VTB:
                // Activate thermistor bias by writing to bq76925.
                // Wait ~32 ms then measure with ADC.

                // Set VTB enable bit in POWER_CTL register
                if (i2c_write(POWER_CTL, 0x0F, &i2c_write_struct)) {
                    error_trap();
                }
                // Delay 32 ms then take measurement
                DELAY_32MS;
                adc_data = ADC_read(INCH_0);

                // Disable VTB
                if (i2c_write(POWER_CTL, 0x0D, &i2c_write_struct)) {
                    error_trap();
                }

            } else if (index == 7) {
                // Measure VIOUT:
                // Measure SENSEP and then calculate result from SENSEN
                // calibration.
                adc_data = ADC_read(INCH_2);
                adc_data = adc_sensen - adc_data;

            } else {
                // Measure VCn via VCOUT (n = 1 to 6):
                // Setup cell MUX in bq76925 to select desired cell input.
                // Wait ~32 ms then measure with ADC.
                if (i2c_write(CELL_CTL, 0x10 | (index - 1), &i2c_write_struct)) {
                    error_trap();
                }
                
                // Delay 32 ms then take measurement
                DELAY_32MS;
                adc_data = ADC_read(INCH_1);        
            }
        }
            
        // Debug code start
        // Toggle the P1.0 GPIO each time through loop       
        P1OUT ^= 0x01;
        // Debug code end
                       
        RESET_WATCHDOG;         // Clear watchdog timer and restart
    }
}

void clock_init (void) {
    // -------------------------------------------------------------------------
    // Clock configuration
    // Configure clock system using factory calibrated values stored in Flash
    // -------------------------------------------------------------------------

    // Ensure that calibration information has not been erased.
    if (CALBC1_8MHZ == 0xFF || CALDCO_8MHZ == 0xFF) {
        // If calibration constants are erased, trap the CPU
        error_trap();
    }   
            
    // Stop watchdog timer
    HALT_WATCHDOG;

    // Set up the clocks as follows:
    //     Digitally-controlled oscillator clock, DCOCLK = 8 MHz
    //     Main system clock, MCLK = DCOCLK
    //     Sub system clock, SMCLK = DCOCLK
    //     Auxiliary clock, ACLK = Internal LP/LF oscillator (12 kHz)
    DCOCTL  = CALDCO_8MHZ;                  // Use calibration constant to set
    BCSCTL1 = CALBC1_8MHZ;                  //   DCOCLK = 8 MHz
    BCSCTL2 = DIVS_0;                       // MCLK = SMCLK = DCOCLK
    BCSCTL3 = LFXT1S_2;                     // Set ACLK source to 12 kHz

    // Set watchdog timer for 8192 cycles of ACLK = 682 ms @ 12 kHz. A PUC will
    // be generated if the watchdog timer is not reset periodically at an 
    // interval < 682 ms.
    RESET_WATCHDOG;
 }
 
void i2c_init (void) {
    // -------------------------------------------------------------------------
    // I2C configuration
    // Configure USCI_B0 module for I2C communications
    // -------------------------------------------------------------------------
    P3SEL |= 0x06;                          // Assign I2C pins to USCI_B0 
                                            //   (UCB0SCL & UCB0SDA)
    UCB0CTL1 |= UCSWRST;                    // Set SW reset, halt I2C operation
    UCB0CTL0 = UCMST + UCMODE_3 + UCSYNC;   // I2C Master, 7-bit address
    UCB0CTL1 = UCSSEL_2 + UCSWRST;          // Use SMCLK, keep SW reset
    UCB0BR0 = 80;                           // fSCL = SMCLK / 80 = 100 kHz 
    UCB0BR1 = 0;                            
    UCB0CTL1 &= ~UCSWRST;                   // Clear SW reset, resume operation
}

int i2c_read(unsigned char reg_addr, i2c_trans *i2c_read) {
    // -------------------------------------------------------------------------
    // I2C read
    // Read data from a bq76925 register to MSP430
    // 
    // Arguments
    //     reg_addr     = Address of bq76925 register (0x00 - 0x1F)
    //     i2c_read     = Pointer to data structure for storing read data
    //
    // Return value     = Error code (0 = no error, 1 = error)
    // -------------------------------------------------------------------------

    // Set up slave address. Each bq76925 register appears as a separate slave
    // address on the I2C bus. The slave address is calculated by shifting the
    // group address and adding the desired register address. The group address
    // is a constant value set in the hardware of the bq76925.
    UCB0I2CSA = (I2C_GROUP_ADDR << 3) + reg_addr;                       
    
    // Set module to receive mode and generate the start. Setting the start bit
    // (UCTXSTT) will generate the start and transmit the slave address.
    UCB0CTL1 &= ~UCTR; 
    UCB0CTL1 |= UCTXSTT;
    
    // Wait for the start bit to clear. The start bit clears after the
    // acknowledge is received. If the slave address was ACK'ed, then reading
    // of the first data byte starts automatically. When receiving only a single
    // byte, the stop bit (UCTXSTP) must be set while the byte is being
    // received. The stop bit must also be set if the slave address was NACK'ed.
    // Therefore, in the case of either an ACK or a NACK, the stop bit is set 
    // immediately after the start bit clears. 
    RESET_WATCHDOG;             // Clear watchdog timer and restart
    while (UCB0CTL1 & UCTXSTT);
    UCB0CTL1 |= UCTXSTP;
    
    // Check the NACK flag (UCNACKIFG). If the NACK flag is clear, then the
    // slave address was ACK'ed and the first data byte is in transit. Otherwise
    // the slave address was NACK'ed.
    if(!(UCB0STAT & UCNACKIFG)) {
        // Slave address was ACK'ed. Wait for the receive flag (UCB0RXIFG) to be
        // set. The receive flag is set following reception of the data byte.
        RESET_WATCHDOG;         // Clear watchdog timer and restart
        while (!(IFG2 & UCB0RXIFG));
        
        // Read data from I2C buffer and store in read transaction data
        // structure. Set error status to 0 to indicate no error.
        (*i2c_read).data  = UCB0RXBUF;
        (*i2c_read).error = 0;
        
    } else {
        // Slave address NACK'ed. Fill read transaction data with 0. Set error
        // status to 1 to indicate error.
        (*i2c_read).data  = 0;
        (*i2c_read).error = 1;  
    }
    
    // Wait for the stop bit to clear. The stop bit clears when the stop is 
    // complete, signaling the end of the transaction.
    RESET_WATCHDOG;             // Clear watchdog timer and restart
    while (UCB0CTL1 & UCTXSTP);

    // Return error status
    return (*i2c_read).error;
}

int i2c_write (unsigned char reg_addr, unsigned char data, i2c_trans *i2c_write) {
    // -------------------------------------------------------------------------
    // I2C write
    // Write data from MSP430 to a bq76925 register
    //
    // Arguments
    //     reg_addr     = Address of bq76925 register (0x00 - 0x1F)
    //     data         = Data to be written to bq76925 register
    //     i2c_write    = Pointer to data structure for storing write data
    //
    // Return value     = Error code (0 = no error, 1 = error)
    // -------------------------------------------------------------------------

    // Set up slave address. Each bq76925 register appears as a separate slave
    // address on the I2C bus. The slave address is calculated by shifting the
    // group address and adding the desired register address. The group address
    // is a constant value set in the hardware of the bq76925.
    UCB0I2CSA = (I2C_GROUP_ADDR << 3) + reg_addr;                       
    
    // Set module to transmit mode and generate the start. Setting the start bit
    // (UCTXSTT) will generate the start and transmit the slave address.
    UCB0CTL1 |= UCTR + UCTXSTT; 

    // Wait for the transmit flag (UCB0TXIFG) to be set. The transmit flag is
    // set when the start is generated and the first data to be transmitted can
    // be written into the transmit buffer (UCB0TXBUF). Loading the transmit
    // buffer clears UCB0TXIFG.
    RESET_WATCHDOG;             // Clear watchdog timer and restart
    while (!(IFG2 & UCB0TXIFG));
    UCB0TXBUF = data;           // Load TX buffer with data

    // Wait for the start bit to clear. The start bit clears after the
    // acknowledge is received. If the slave address was ACK'ed, then sending 
    // of the first data byte starts automatically. When transmitting only a
    // single byte, the stop bit (UCTXSTP) must be set while the byte is being
    // transmitted. The stop bit must also be set if the slave address was
    // NACK'ed. Therefore, in the case of either an ACK or a NACK, the stop bit
    // is set immediately after the start bit clears. 
    RESET_WATCHDOG;             // Clear watchdog timer and restart
    while (UCB0CTL1 & UCTXSTT);
    UCB0CTL1 |= UCTXSTP;

    // Check the NACK flag (UCNACKIFG). If the NACK flag is clear, then the
    // slave address was ACK'ed and the first data byte is in transit. Otherwise
    // the slave address was NACK'ed.
    if(!(UCB0STAT & UCNACKIFG)) {
        // If the data byte was sent, the transmit flag is set and needs
        // to be cleared.
        IFG2 &= ~UCB0TXIFG; 
    }   

    // Wait for the stop bit to clear. The stop bit clears when the stop is 
    // complete, signaling the end of the transaction.
    RESET_WATCHDOG;             // Clear watchdog timer and restart
    while(UCB0CTL1 & UCTXSTP);
        
    // Check the NACK flag. The NACK flag will indicate the status of the last
    // acknowledge sent from the bq76925. In the case where the slave address
    // was NACK'ed, the flag will be set. In the case where the slave address
    // was ACK'ed, the NACK flag will indicate the acknowledge status of the
    // transmitted data byte.
    if(!(UCB0STAT & UCNACKIFG)) {
        // Data byte ACK'ed. Store transmitted data in the write transaction
        // data structure. Set error status to 0 to indicate no error.
        (*i2c_write).data  = data;
        (*i2c_write).error = 0;
        
    } else {
        // Slave address or data byte was NACK'ed. Fill write transaction data
        // with 0. Set error status to 1 to indicate error.
        (*i2c_write).data =  0;
        (*i2c_write).error = 1;
    }
    
    // Return error status
    return (*i2c_write).error;
}

void ADC_init (void) {
    // -------------------------------------------------------------------------
    // ADC configuration
    // Configure ADC10 module for A/D conversions
    // -------------------------------------------------------------------------
    ADC10AE0 |= 0x07;                   // A0, A1 and A2 enabled as analog inputs

    // Control Register 0 Configuration
    // The internal reference is not used
    ADC10CTL0 = SREF_2 +                // Use external reference (VeREF+ to VSS)
                ADC10SHT_3 +            // Sample-and-hold time = 64 x ADC10CLKs
                                        //   = 64 us
                ADC10ON;                // Turn ADC10 module on

    // Control Register 1 Configuration
    // Single channel, single conversion
    // Data format is straight binary
    // No inversion
    ADC10CTL1 = INCH_0 +                // Default input channel is A0 (VIOUT)
                SHS_0 +                 // Sample-and-hold source is start
                                        //   conversion bit (ADC10SC)
                ADC10SSEL_3 +           // ADC10CLK source is SMCLK (8 MHz)
                ADC10DIV_7;             // ADC10CLK = SMCLK / 8 = 1 MHz
}

unsigned int ADC_read (unsigned int ch_selected) {
    // -------------------------------------------------------------------------
    // ADC read
    // Perform ADC conversion and return results.
    //
    // Arguments
    //     ch_selected  = Analog input channel for conversion 
    //
    // Return value     = A/D result
    // -------------------------------------------------------------------------

    ADC10CTL0 &= ~ENC;                      // Disable ADC10 conversion
    
    // Set the analog input channel for conversion
    ADC10CTL1 &= (~INCH_15);                // Clear old channel
    ADC10CTL1 |= ch_selected;               // Set new channel
    
    // Start the conversion
    ADC10CTL0 &= ~ADC10IFG;                 // Clear the ADC interrupt flag
    ADC10CTL0 |= ENC + ADC10SC;             // Enable and start conversion
    
    // Wait for interrupt flag to indicate a completed A/D conversion
    RESET_WATCHDOG;             // Clear watchdog timer and restart
    while (!(ADC10CTL0 & ADC10IFG));

    // Return conversion result
    return ADC10MEM;
}

void error_trap (void) {
    // -------------------------------------------------------------------------
    // Error trap
    // This error handling routine is simply an infinite loop. The device will
    // reset once the watchdog timer expires. In an actual application this
    // would be filled in with code appropriate for the application.
    // -------------------------------------------------------------------------
    while (1);
}
