/*
 *  MSP430F2122 Demo - ADC10, I2C Sample Code for bq76925 Analog Front End (AFE)
 *
 *  Description: This sample code demonstrates I2C communication, A/D
 *  conversion, correction factor usage and fault limit comparison in a system
 *  comprising the bq76925 and an MSP430.
 *
 *  The code demonstrates how to read and write the bq76925 registers and 
 *  perform A/D conversions to obtain current, voltage and temperature readings
 *  from the VIOUT, VCOUT and VTB pins of the bq76925.
 *
 *  The code further demonstrates the application of correction factors to
 *  minimize errors in the measurement and compares the corrected measurements
 *  against thresholds for fault detection.
 *
 *  The following connections between the bq76925 and MSP430 are assumed:
 *
 *  | bq76925 pin name | MSP430 pin name |
 *  | ---------------- | --------------- |
 *  |        SCL       |     UCB0SCL     |
 *  |        SDA       |     UCB0SDA     |
 *  |       VREF       |      VREF+      |
 *  |        VTB       |       A0*       |
 *  |       VCOUT      |       A1        |
 *  |       VIOUT      |       A2        |
 *
 *  * VTB is not connected directly to A0. It is connected through a series
 *  resistance that makes up part of an NTC thermistor network as follows:
 *
 *  VTB -----[ 10K ]-----*----- A0
 *                       |
 *                       |
 *                     103AT
 *                       |
 *                       |
 *                      VSS
 *
 *  The GPIO's of the MSP430 are assigned as follows:
 * 
 *  | GPIO | Description                              |
 *  | ---- | ---------------------------------------- |
 *  | P1.0 | 0 = Over temperature fault, 1 = no fault |
 *  | P1.1 | 0 = Over current fault, 1 = no fault     |
 *  | P1.2 | 0 = Under voltage fault, 1 = no fault    |
 *  | P1.3 | 0 = Over voltage fault, 1 = no fault     |
 *
 *  Michael Lu / J. Ross
 *  Texas Instruments Inc.
 *  Rev 1.0, 2 September 2011
 *  Built with CCS Version: 4.5.0
 *
 *  Revision history:
 *
 *      Date        Rev     Description
 *  ------------    ---     ----------------------------------------------
 *   2 Sept 2011    1.0     Initial release
 *
 * Copyright (C) 2011 Texas Instruments Incorporated - http://www.ti.com/ 
 * 
 * 
 *  Redistribution and use in source and binary forms, with or without 
 *  modification, are permitted provided that the following conditions 
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the   
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
*/

// Include files
// =============================================================================
#include <intrinsics.h>
#include "msp430x21x2.h"

// =============================================================================
// Define macros for watchdog timer
// =============================================================================
#define HALT_WATCHDOG   WDTCTL = WDTPW + WDTHOLD
#define RESET_WATCHDOG  WDTCTL = WDTPW + WDTCNTCL + WDTSSEL + WDTIS0

// =============================================================================
// Define time delays using compiler intrinsics
// =============================================================================
#define DELAY_32MS      __delay_cycles(256000)      // 32 ms (8 MHz clock)
#define DELAY_256MS     __delay_cycles(256000*8)    // 256 ms (8 MHz clock)

// =============================================================================
// Define bq76925 combined register addresses
//
// Each register in the bq76925 appears as a separate device slave address on 
// the I2C bus. The slave address is a combination of the bq76925 group address
// and the absolute register address. The slave address is calculated by
// shifting the group address left by 3 and adding the absolute register
// address. The group address is a constant value set in the hardware of the
// bq76925 and is equal to 4.
// =============================================================================
#define I2C_GROUP_ADDR  0x4
#define STATUS          (I2C_GROUP_ADDR << 3) + 0x00    // Read & Write
#define CELL_CTL        (I2C_GROUP_ADDR << 3) + 0x01    // Read & Write
#define BAL_CTL         (I2C_GROUP_ADDR << 3) + 0x02    // Read & Write
#define CONFIG_1        (I2C_GROUP_ADDR << 3) + 0x03    // Read & Write
#define CONFIG_2        (I2C_GROUP_ADDR << 3) + 0x04    // Read & Write
#define POWER_CTL       (I2C_GROUP_ADDR << 3) + 0x05    // Read & Write
#define CHIP_ID         (I2C_GROUP_ADDR << 3) + 0x07    // Read Only
#define VREF_CAL        (I2C_GROUP_ADDR << 3) + 0x10    // Read Only
#define VC1_CAL         (I2C_GROUP_ADDR << 3) + 0x11    // Read Only
#define VC2_CAL         (I2C_GROUP_ADDR << 3) + 0x12    // Read Only
#define VC3_CAL         (I2C_GROUP_ADDR << 3) + 0x13    // Read Only
#define VC4_CAL         (I2C_GROUP_ADDR << 3) + 0x14    // Read Only
#define VC5_CAL         (I2C_GROUP_ADDR << 3) + 0x15    // Read Only
#define VC6_CAL         (I2C_GROUP_ADDR << 3) + 0x16    // Read Only
#define VC_CAL_EXT_1    (I2C_GROUP_ADDR << 3) + 0x17    // Read Only
#define VC_CAL_EXT_2    (I2C_GROUP_ADDR << 3) + 0x18    // Read Only
#define VREF_CAL_EXT    (I2C_GROUP_ADDR << 3) + 0x1B    // Read Only

// =============================================================================
// Define references for measurement calculations
//
// VREF_NOM : Nominal reference voltage. Set by writing REF_SEL in the CONFIG_2
//            register (REF_SEL = 0, 1 => VREF_NOM = 1.5, 3.0).
// GVCOUT   : Nominal cell voltage gain. Also set by writing REF_SEL in the
//            CONFIG_2 register (REF_SEL = 0, 1 => GVCOUT = 0.3, 0.6)  .
// GVSENSE  : Current amplifier gain. Set by writing the I_GAIN bit in the
//            CONFIG_1 register (I_GAIN = 0, 1 => GVSENSE = 4, 8).
// FS_CNT   : ADC full-scale count. For an 'n' bit ADC, FS_CNT = pow(2,n) - 1.
// =============================================================================
#define VREF_NOM        3
#define GVCOUT          0.6
#define GVSENSE         4
#define FS_CNT          1023L   // 10-bit ADC

// =============================================================================
// Define fault thresholds
//
// The fault thresholds are transformed using function like macros. The trans-
// formation reduces code size by avoiding floating point arithmetic and
// division.
//
// VC_TRANS(x)  : Transforms voltage threshold. 'x' is in Volts.
// VI_TRANS(x)  : Transforms current threshold. 'x' is in Volts.
// VT_TRANS(x)  : Transforms temperature threshold. 'x' is in Volts.
// =============================================================================
#define VC_TRANS(x)     ((unsigned int) (x * GVCOUT * FS_CNT * 1e6 / 65536))
#define VI_TRANS(x)     ((unsigned int) (x * GVSENSE * FS_CNT * 1e3 / 256))
#define VT_TRANS(x)     ((unsigned int) (x * FS_CNT * 1e3 / 256))

#define OV_THRESH       4.250   // Over voltage threshold in Volts
#define UV_THRESH       2.500   // Under voltage threshold in Volts
#define OC_THRESH       0.125   // Over current threshold in Volts
#define OT_THRESH       0.700   // Over temperature threshold in Volts

// =============================================================================
// Type definitions
// =============================================================================
// Data structure for I2C transactions
typedef struct {
    unsigned char data;
    unsigned char error;              
} i2c_trans;

// =============================================================================
// Function prototypes
// =============================================================================
void clock_init (void);

void i2c_init (void);

int i2c_read (unsigned char reg_addr, i2c_trans *i2c_read);

int i2c_write_and_verify (unsigned char reg_addr, unsigned char data, i2c_trans *i2c_write);

void ADC_init (void);

unsigned int ADC_read (unsigned int ch_selected);

void error_trap (void);

// =============================================================================
// Main loop
// =============================================================================
int main (void) {
    
    // Declare structures to hold latest read and write I2C transaction
    i2c_trans i2c_read_struct, i2c_write_struct;
    
    // Declare arrays that will hold offset and gain correction factors 
    // for VREF and VC1-VC6. Indices are arranged in the following order:
    // 
    //     | Index | Channel |
    //     | ----- | ------- |
    //     |   0   |   VREF  |
    //     |   1   |   VC1   |
    //     |   2   |   VC2   |
    //     |   3   |   VC3   |
    //     |   4   |   VC4   |
    //     |   5   |   VC5   |
    //     |   6   |   VC6   |
    //
    signed char offset_corr[7];
    signed char gain_corr[7];
    
    unsigned int vref_corr;         // Corrected VREF x 1000
    
    volatile unsigned int adc_data; // ADC data from conversion
    unsigned int adc_sensen;        // Store value of SENSEN measurement

    unsigned int index;             // General purpose index
    unsigned char ov_flag,uv_flag;  // Fault flags
    
    // =========================================================================
    // MSP430 configuration setup
    //
    clock_init();                   // Initialize clock system
    i2c_init();                     // Initialize I2C module
    ADC_init();                     // Initialize ADC10 module

    // Initialize fault indicators
    // The P1.0, P1.1, P1.2 and P1.3 GPIO are used to signal a fault.
    P1DIR = 0x0F;                   // Set GPIO's as output
    P1OUT = ~(0x00);                // Initialize to all 1's
    
    // =========================================================================
    // bq76925 configuration setup
    //
    // Select SENSEN measurement (I_AMP_CAL = 0) and gain of 4 (I_GAIN = 0)
    if (i2c_write_and_verify(CONFIG_1, 0x00, &i2c_write_struct)) error_trap();
    // Set VREF_NOM = 3.0 V and GVCOUT = 0.6 (REF_SEL = 1)
    if (i2c_write_and_verify(CONFIG_2, 0x01, &i2c_write_struct)) error_trap();
    // Enable VREF, VCOUT and VIOUT.
    if (i2c_write_and_verify(POWER_CTL, 0x0D, &i2c_write_struct)) error_trap();
   
    // =========================================================================
    // Assemble correction factors from registers
    //
    // All correction factors are 5 bits, except for VREF offset correction,
    // which is 6-bits.
    
    // Get the 4 lsb's of the VREF and VC1-VC6 correction factors by reading
    // the 7 registers starting with VREF_CAL.
    for (index = 0; index < 7; index++) {
        if (i2c_read( VREF_CAL | index, &i2c_read_struct)) error_trap();
        offset_corr[index] = i2c_read_struct.data >> 4;
        gain_corr[index]   = i2c_read_struct.data & 0x0F;
    }
    
// Un-comment the next line ONLY if the bq76925 was distributed prior to
// product release.    
//#define PRE_RELEASE_SAMPLES

#ifdef PRE_RELEASE_SAMPLES
    // Prior to product release, some samples and EVM's were distibuted that
    // had only 4-bit correction factors. These simply need a sign-extension.
    for (index = 0; index < 7; index++) {
        offset_corr[index] |= (offset_corr[index] ^ 0x08) - 0x08;
        gain_corr[index]   |= (gain_corr[index] ^ 0x08) - 0x08;
    }
#else
    // Production product has 5 and 6 bit correction factors. These need
    // to be assembled by reading the msb's from a separate register and then
    // sign extended.
    
    // Get the 2 msb for offset and 1 msb for gain correction of VREF
    if (i2c_read(VREF_CAL_EXT, &i2c_read_struct)) error_trap();
    // Shift in msb and sign extend
    offset_corr[0] |= (((i2c_read_struct.data & 0x06) << 3) ^ 0x20) - 0x20;
    gain_corr[0]   |= (((i2c_read_struct.data & 0x01) << 4) ^ 0x10) - 0x10;
        
    // Get the msb for offset and gain corrections for VC1 and VC2
    if (i2c_read(VC_CAL_EXT_1, &i2c_read_struct)) error_trap();
    // Shift in msb and sign extend
    offset_corr[1] |= (((i2c_read_struct.data & 0x80) >> 3) ^ 0x10) - 0x10;
    gain_corr[1]   |= (((i2c_read_struct.data & 0x40) >> 2) ^ 0x10) - 0x10;
        
    offset_corr[2] |= (((i2c_read_struct.data & 0x20) >> 1) ^ 0x10) - 0x10;
    gain_corr[2]   |= (((i2c_read_struct.data & 0x10)) ^ 0x10) - 0x10;
            
    // Get the msb for offset and gain corrections for VC3, VC4, VC5, VC6
    if (i2c_read(VC_CAL_EXT_2, &i2c_read_struct)) error_trap();
    // Shift in msb and sign extend
    offset_corr[3] |= (((i2c_read_struct.data & 0x80) >> 3) ^ 0x10) - 0x10;
    gain_corr[3]   |= (((i2c_read_struct.data & 0x40) >> 2) ^ 0x10) - 0x10;
    
    offset_corr[4] |= (((i2c_read_struct.data & 0x20) >> 1) ^ 0x10) - 0x10;
    gain_corr[4]   |= (((i2c_read_struct.data & 0x10)) ^ 0x10) - 0x10;
    
    offset_corr[5] |= (((i2c_read_struct.data & 0x08) << 1) ^ 0x10) - 0x10;
    gain_corr[5]   |= (((i2c_read_struct.data & 0x04) << 2) ^ 0x10) - 0x10;
    
    offset_corr[6] |= (((i2c_read_struct.data & 0x02) << 3) ^ 0x10) - 0x10;
    gain_corr[6]   |= (((i2c_read_struct.data & 0x01) << 4) ^ 0x10) - 0x10;
#endif    
        
    // =========================================================================
    // Correction Factor Application to VREF
    // Use VREF Gain and Offset correction factors to find corrected (actual)
    // VREF for future calculations. Multiply through by 1000 to keep results as
    // integers: vref_corr = 1000 * VREF, in Volts.
    vref_corr = VREF_NOM * (1000L + gain_corr[0]) + offset_corr[0];

    // =========================================================================
    // Continuous measurement of VIOUT, VCOUT, and VTB.
    //
    // Eight AFE channels are defined as shown in the table below. The channels
    // are cycled through continuously to measure current via VIOUT, all cell
    // voltages (VC1 - VC6) via VCOUT, and temperature from a thermistor biased
    // from VTB.
    //
    // |             |  AFE Cell  |  Measure   | MSP430 Analog |
    // | AFE Channel | MUX Select | AFE Output | Input Select  |
    // | ----------- | ---------- | ---------- | ------------- |
    // |      0      |    VSS     |    VTB     |      A0       |
    // |      1      |    VC1     |   VCOUT    |      A1       |
    // |      2      |    VC2     |   VCOUT    |      A1       |
    // |      3      |    VC3     |   VCOUT    |      A1       |
    // |      4      |    VC4     |   VCOUT    |      A1       |
    // |      5      |    VC5     |   VCOUT    |      A1       |
    // |      6      |    VC6     |   VCOUT    |      A1       |
    // |      7      |    VSS     |   VIOUT    |      A2       |
    //

    // Before starting loop, SENSEN is measured. When the current amplifier is
    // turned on, it must charge up to the zero current level (~ 2V). If the
    // output filter is large, it may take awhile for VIOUT to reach a steady
    // value. Here, we delay ~2 sec before measurement to accomodate a very 
    // large filter. Smaller delays can be used for smaller filters.
    for (index = 0; index < 8 ; index++) {
        RESET_WATCHDOG;
        DELAY_256MS;
    }
    
    // Measure SENSEN ADC value
    adc_sensen = ADC_read(INCH_2);

    // Switch bq76925 to SENSEP for all other current current measurements
    if (i2c_write_and_verify(CONFIG_1, 0x04, &i2c_write_struct)) error_trap();

    while (1) {

        // Measure temperature
        // -------------------
        // Activate thermistor bias on VTB pin by setting enable bit in
        // bq76925 POWER_CTL register.
        if (i2c_write_and_verify(POWER_CTL, 0x0F, &i2c_write_struct)) error_trap();

        // Delay 32 ms before taking measurement. Delay takes into account the
        // time required to charge up any capacitance in thermistor network and
        // should be adjusted according to circuit values. 
        DELAY_32MS;

        // Multiply ADC count by vref_corr and divide by 256 to give a
        // 16-bit result 
        adc_data = ADC_read(INCH_0) * (long) vref_corr >> 8;

        // Disable thermistor bias on VTB pin
        if (i2c_write_and_verify(POWER_CTL, 0x0D, &i2c_write_struct)) error_trap();
   
        // Determine if fault threshold is violated by comparing the adc_data
        // output to the over temperature voltage threshold
        if (adc_data < VT_TRANS(OT_THRESH)) {
            // Clear P1.0 if over temperature fault detected
            P1OUT &= 0xFE;    
        } else {
            // Set P1.0 if fault is not detected
            P1OUT |= 0x01;
        } 
        
        // Measure current
        // ---------------
        // Measure SENSEP via VIOUT and then calculate result using SENSEN
        // measurement taken above.
        adc_data = ADC_read(INCH_2);
        
        // Check for discharge current
        if (adc_data <= adc_sensen) {
            // If current is discharge, take difference of SENSEN and SENSEP,
            // multiply ADC count by vref_corr and divide by 256 to give 16-bit
            // result.
            adc_data = ((long) (adc_sensen - adc_data) * vref_corr ) >> 8;
        
            // Determine if fault threshold is violated by comparing
            // the result to the over current threshold voltage
            if (adc_data > VI_TRANS(OC_THRESH)) {
                // Clear P1.1 if over current fault detected
                P1OUT &= 0xFD;   
            } else {
                // Set P1.1 if fault is not detected
                P1OUT |= 0x02;
            } 
        }

        // Measure cell voltages
        // ---------------------
        // Initialize fault flags
        ov_flag = 0;
        uv_flag = 0;
        
        // Measure VCn via VCOUT for n = 1 to 6
        // Cycle through all the cell voltage channels
        for (index = 0; index < 6; index++) {
            // Setup cell MUX in bq76925 to select desired cell input.
            if (i2c_write_and_verify(CELL_CTL, 0x10 | index, &i2c_write_struct)) error_trap();
            
            // Delay 32 ms before taking measurement. Delay takes into account
            // the time required to charge or discharge any capacitance in the
            // output filter on VCOUT and should be adjusted according to 
            // the max voltage swing and circuit values.
            DELAY_32MS;
            
            // Corrected VCOUT voltage = 
            //
            //       ADC Count              offset_corr          gain_corr
            //  ( ---------------- x VREF + ----------- ) x (1 + ---------)
            //    Full-scale count             1000                1000
            //
            // Multiply through by Full-scale count x 1E6. Recall that
            // 1000 x VREF = vref_corr. Divide by 65536 to get 16-bit result.
            adc_data = ((long) ADC_read(INCH_1) * vref_corr + FS_CNT * offset_corr[index+1]) \
                       * (1000L + gain_corr[index+1]) >> 16;

            // Check for under voltage
            if (adc_data < VC_TRANS(UV_THRESH)) uv_flag = 1;
            // Check for over voltage
            if (adc_data > VC_TRANS(OV_THRESH)) ov_flag = 1;
        }
        
        // After checking all cells, set the fault indicators
        if (uv_flag) {
            // Clear P1.2 if UV fault detected
            P1OUT &= 0xFB; 
        } else {
            // Set P1.2 if fault is not detected
            P1OUT |= 0x04;
        }        
         
        if (ov_flag) {
            // Clear P1.3 if OV fault detected
            P1OUT &= 0xF7; 
        } else {
            // Set P1.3 if fault is not detected
            P1OUT |= 0x08;
        }        

        RESET_WATCHDOG;         // Clear watchdog timer and restart
    }
}

void clock_init (void) {
    // =========================================================================
    // Clock Configuration
    // Configure clock system using factory calibrated values stored in Flash
    // =========================================================================

    // Ensure that calibration information has not been erased.
    if (CALBC1_8MHZ == 0xFF || CALDCO_8MHZ == 0xFF) {
        // If calibration constants are erased, trap the CPU
        error_trap();
    }   
            
    // Stop watchdog timer
    HALT_WATCHDOG;

    // Set up the clocks as follows:
    //     Digitally-controlled oscillator clock, DCOCLK = 8 MHz
    //     Main system clock, MCLK = DCOCLK
    //     Sub system clock, SMCLK = DCOCLK
    //     Auxiliary clock, ACLK = Internal LP/LF oscillator (12 kHz)
    DCOCTL  = CALDCO_8MHZ;                  // Use calibration constant to set
    BCSCTL1 = CALBC1_8MHZ;                  //      DCOCLK = 8 MHz
    BCSCTL2 = DIVS_0;                       // MCLK = SMCLK = DCOCLK
    BCSCTL3 = LFXT1S_2;                     // Set ACLK source to 12 kHz

    // Set watchdog timer for 8192 cycles of ACLK = 682 ms @ 12 kHz. A PUC will
    // be generated if the watchdog timer is not reset periodically at an 
    // interval < 682 ms.
    RESET_WATCHDOG;
 }
 
void i2c_init (void) {
    // =========================================================================
    // I2C Configuration
    // Configure USCI_B0 module for I2C communications
    // =========================================================================
    P3SEL |= 0x06;                          // Assign I2C pins to USCI_B0 
                                            //   (UCB0SCL & UCB0SDA)
    UCB0CTL1 |= UCSWRST;                    // Set SW reset, halt I2C operation
    UCB0CTL0 = UCMST + UCMODE_3 + UCSYNC;   // I2C Master, 7-bit address
    UCB0CTL1 = UCSSEL_2 + UCSWRST;          // Use SMCLK, keep SW reset
    UCB0BR0 = 80;                           // fSCL = SMCLK / 80 = 100 kHz 
    UCB0BR1 = 0;                            
    UCB0CTL1 &= ~UCSWRST;                   // Clear SW reset, resume operation
}

int i2c_read(unsigned char reg_addr, i2c_trans *i2c_read) {
    // =========================================================================
    // I2C Read
    // Read data from a bq76925 register to MSP430
    // 
    // Arguments
    //     reg_addr     = Combined address of bq76925 register formed by
    //                    shifting the group address left by 3 and adding the
    //                    base register address
    //     i2c_read     = Pointer to data structure for storing read data
    //
    // Return value     = Error code:
    //                        0 = no error
    //                        1 = error during transaction 
    // =========================================================================

    // Set up slave address. Each bq76925 register appears as a separate slave
    // address on the I2C bus.
    UCB0I2CSA = reg_addr;                       
    
    // Set module to receive mode and generate the start. Setting the start bit
    // (UCTXSTT) will generate the start and transmit the slave address.
    UCB0CTL1 &= (~UCTR); 
    UCB0CTL1 |= UCTXSTT;
    
    // Wait for the start bit to clear. The start bit clears after the
    // acknowledge is received. If the slave address was ACK'ed, then reading
    // of the first data byte starts automatically. When receiving only a single
    // byte, the stop bit (UCTXSTP) must be set while the byte is being
    // received. The stop bit must also be set if the slave address was NACK'ed.
    // Therefore, in the case of either an ACK or a NACK, the stop bit is set 
    // immediately after the start bit clears. 
    RESET_WATCHDOG;             // Clear watchdog timer and restart
    while (UCB0CTL1 & UCTXSTT);
    UCB0CTL1 |= UCTXSTP;
    
    // Check the NACK flag (UCNACKIFG). If the NACK flag is clear, then the
    // slave address was ACK'ed and the first data byte is in transit. Otherwise
    // the slave address was NACK'ed.
    if(!(UCB0STAT & UCNACKIFG)) {
        // Slave address was ACK'ed. Wait for the receive flag (UCB0RXIFG) to be
        // set. The receive flag is set following reception of the data byte.
        RESET_WATCHDOG;         // Clear watchdog timer and restart
        while (!(IFG2 & UCB0RXIFG));
        
        // Read data from I2C buffer and store in read transaction data
        // structure. Set error status to 0 to indicate no error.
        (*i2c_read).data  = UCB0RXBUF;
        (*i2c_read).error = 0;
        
    } else {
        // Slave address NACK'ed. Fill read transaction data with 0. Set error
        // status to 1 to indicate error.
        (*i2c_read).data  = 0;
        (*i2c_read).error = 1;  
    }
    
    // Wait for the stop bit to clear. The stop bit clears when the stop is 
    // complete, signaling the end of the transaction.
    RESET_WATCHDOG;             // Clear watchdog timer and restart
    while (UCB0CTL1 & UCTXSTP);

    // Return error status
    return (*i2c_read).error;
}

int i2c_write_and_verify (unsigned char reg_addr, unsigned char data, i2c_trans *i2c_write) {
    // =========================================================================
    // I2C Write & Verify
    // Write data from MSP430 to a bq76925 register then read-back to verify.
    // This function performs two I2C transactions. The first transaction writes
    // 'data' to the address 'reg_addr'. The second transaction reads the
    // address 'reg_addr' and compares the read data against the write data.
    //
    // Arguments
    //     reg_addr     = Combined address of bq76925 register formed by
    //                    shifting the group address left by 3 and adding the
    //                    base register address
    //     data         = Data to be written to bq76925 register
    //     i2c_write    = Pointer to data structure for storing write data
    //
    // Return value     = Error code:
    //                        0 = no error
    //                        1 = error during write transaction 
    //                        2 = error during read transaction
    //                        3 = write and read data mismatch
    // =========================================================================
    
    // Declare local read transaction struct for data read-back
    i2c_trans i2c_read_struct;
    
    // Set up slave address. Each bq76925 register appears as a separate slave
    // address on the I2C bus.
    UCB0I2CSA = reg_addr;                       
    
    // Set module to transmit mode and generate the start. Setting the start bit
    // (UCTXSTT) will generate the start and transmit the slave address.
    UCB0CTL1 |= UCTR + UCTXSTT; 

    // Wait for the transmit flag (UCB0TXIFG) to be set. The transmit flag is
    // set when the start is generated and the first data to be transmitted can
    // be written into the transmit buffer (UCB0TXBUF). Loading the transmit
    // buffer clears UCB0TXIFG.
    RESET_WATCHDOG;             // Clear watchdog timer and restart
    while (!(IFG2 & UCB0TXIFG));
    UCB0TXBUF = data;           // Load TX buffer with data

    // Wait for the start bit to clear. The start bit clears after the
    // acknowledge is received. If the slave address was ACK'ed, then sending 
    // of the first data byte starts automatically. When transmitting only a
    // single byte, the stop bit (UCTXSTP) must be set while the byte is being
    // transmitted. The stop bit must also be set if the slave address was
    // NACK'ed. Therefore, in the case of either an ACK or a NACK, the stop bit
    // is set immediately after the start bit clears. 
    RESET_WATCHDOG;             // Clear watchdog timer and restart
    while (UCB0CTL1 & UCTXSTT);
    UCB0CTL1 |= UCTXSTP;

    // Check the NACK flag (UCNACKIFG). If the NACK flag is clear, then the
    // slave address was ACK'ed and the first data byte is in transit. Otherwise
    // the slave address was NACK'ed.
    if(!(UCB0STAT & UCNACKIFG)) {
        // If the data byte was sent, the transmit flag is set and needs
        // to be cleared.
        IFG2 &= ~UCB0TXIFG; 
    }   

    // Wait for the stop bit to clear. The stop bit clears when the stop is 
    // complete, signaling the end of the transaction.
    RESET_WATCHDOG;             // Clear watchdog timer and restart
    while(UCB0CTL1 & UCTXSTP);
        
    // Check the NACK flag. The NACK flag will indicate the status of the last
    // acknowledge sent from the bq76925. In the case where the slave address
    // was NACK'ed, the flag will be set. In the case where the slave address
    // was ACK'ed, the NACK flag will indicate the acknowledge status of the
    // transmitted data byte.
    if(!(UCB0STAT & UCNACKIFG)) {
        // Data byte ACK'ed so the data should have been successfully written
        // to the register. But read back the register data to verify.
        if (i2c_read(reg_addr, &i2c_read_struct)) {
            // Error during read-back transaction. Fill write transaction data
            // with 0. Set error status to indicate error during read-back.
            (*i2c_write).data =  0;
            (*i2c_write).error = 2;
        } else {
            // Read-back completed without an error. Fill write transaction data
            // with read-back data.
            (*i2c_write).data = i2c_read_struct.data;
            
            // Check that the read-back data matches the write data
            if (i2c_read_struct.data != data) {
                // Read data does not match write data. Set error status to
                // indicate data mismatch.    
                (*i2c_write).error = 3;
            } else {
                // Set error status to indicate no error.
                (*i2c_write).error = 0;
            }
        }
        
    } else {
        // Slave address or data byte was NACK'ed. Fill write transaction data
        // with 0. Set error status to 1 to indicate error during write.
        (*i2c_write).data =  0;
        (*i2c_write).error = 1;
    }
    
    // Return error status
    return (*i2c_write).error;
}

void ADC_init (void) {
    // =========================================================================
    // ADC Configuration
    // Configure ADC10 module for A/D conversions
    // =========================================================================
    ADC10AE0 |= 0x07;           // A0, A1 and A2 enabled as analog inputs

    // Control Register 0 Configuration
    // The internal reference is not used
    ADC10CTL0 = SREF_2 +        // Use external reference (VeREF+ to VSS)
                ADC10SHT_3 +    // Sample-and-hold time = 64 x ADC10CLKs = 64 us
                ADC10ON;        // Turn ADC10 module on

    // Control Register 1 Configuration
    // Single channel, single conversion
    // Data format is straight binary
    // No inversion
    ADC10CTL1 = INCH_0 +        // Default input channel is A0 (VIOUT)
                SHS_0 +         // Sample-and-hold source is start
                                //   conversion bit (ADC10SC)
                ADC10SSEL_3 +   // ADC10CLK source is SMCLK (8 MHz)
                ADC10DIV_7;     // ADC10CLK = SMCLK / 8 = 1 MHz
}

unsigned int ADC_read (unsigned int ch_selected) {
    // =========================================================================
    // ADC Read
    // Perform ADC conversion and return results.
    //
    // Arguments
    //     ch_selected  = Analog input channel for conversion 
    //
    // Return value     = A/D result
    // =========================================================================
    ADC10CTL0 &= ~ENC;                  // Disable ADC10 conversion
    
    // Set the analog input channel for conversion
    ADC10CTL1 &= (~INCH_15);            // Clear old channel
    ADC10CTL1 |= ch_selected;           // Set new channel
    
    // Start the conversion
    ADC10CTL0 &= (~ADC10IFG);           // Clear the ADC interrupt flag
    ADC10CTL0 |= ENC + ADC10SC;         // Enable and start conversion
    
    // Wait for interrupt flag to indicate a completed A/D conversion
    RESET_WATCHDOG;                     // Clear watchdog timer and restart
    while (!(ADC10CTL0 & ADC10IFG));

    // Return conversion result
    return ADC10MEM;
}

void error_trap (void) {
    // =========================================================================
    // Error Trap
    // This error handling routine is simply an infinite loop. The device will
    // reset once the watchdog timer expires. In an actual application this
    // would be filled in with code appropriate for the application.
    // =========================================================================
    while (1);
}
